@@ -3,19 +3,21 @@ from __future__ import unicode_literals
 from operator import attrgetter
 
 from django import VERSION
-from django.contrib.contenttypes.models import ContentType
 from django.conf import settings
+from django.contrib.contenttypes.models import ContentType
 from django.db import models, router
+from django.db.models import signals
 from django.db.models.fields import Field
-from django.db.models.fields.related import (add_lazy_relation, ManyToManyRel,
-                                             OneToOneRel, RelatedField)
+from django.db.models.fields.related import (ManyToManyRel, OneToOneRel,
+                                             RelatedField)
 from django.utils import six
 from django.utils.text import capfirst
 from django.utils.translation import ugettext_lazy as _
 
 from taggit.forms import TagField
 from taggit.models import CommonGenericTaggedItemBase, TaggedItem
-from taggit.utils import _get_field, require_instance_manager
+from taggit.utils import (_get_field, _related_model, _remote_field,
+                          require_instance_manager)
 
 if VERSION < (1, 8):
     # related.py was removed in Django 1.8
@@ -28,6 +30,13 @@ if VERSION < (1, 8):
 else:
     RelatedObject = None
 
+
+if VERSION >= (1, 9):
+    from django.db.models.fields.related import lazy_related_operation
+else:
+    from django.db.models.fields.related import add_lazy_relation
+
+
 try:
     from django.contrib.contenttypes.fields import GenericRelation
 except ImportError:  # django < 1.7
@@ -43,7 +52,7 @@ except ImportError:  # Django < 1.8
 
 
 def _model_name(model):
-    if VERSION < (1, 7):
+    if VERSION < (1, 6):
         return model._meta.module_name
     else:
         return model._meta.model_name
@@ -83,6 +92,10 @@ class ExtraJoinRestriction(object):
         self.content_types = content_types
 
     def as_sql(self, qn, connection):
+        # qn changed from a quoting function to be a compiler object in 1.8,
+        # which has a quote function
+        if VERSION >= (1, 8):
+            qn = qn.quote_name_unless_alias
         if len(self.content_types) == 1:
             extra_where = "%s.%s = %%s" % (qn(self.alias), qn(self.col))
         else:
@@ -154,16 +167,54 @@ class _TaggableManager(models.Manager):
 
     @require_instance_manager
     def add(self, *tags):
+        db = router.db_for_write(self.through, instance=self.instance)
+
+        tag_objs = self._to_tag_model_instances(tags)
+        new_ids = set(t.pk for t in tag_objs)
+
+        # NOTE: can we hardcode 'tag_id' here or should the column name be got
+        # dynamically from somewhere?
+        vals = (self.through._default_manager.using(db)
+                .values_list('tag_id', flat=True)
+                .filter(**self._lookup_kwargs()))
+
+        new_ids = new_ids - set(vals)
+
+        signals.m2m_changed.send(
+            sender=self.through, action="pre_add",
+            instance=self.instance, reverse=False,
+            model=self.through.tag_model(), pk_set=new_ids, using=db,
+        )
+
+        for tag in tag_objs:
+            self.through._default_manager.using(db).get_or_create(
+                tag=tag, **self._lookup_kwargs())
+
+        signals.m2m_changed.send(
+            sender=self.through, action="post_add",
+            instance=self.instance, reverse=False,
+            model=self.through.tag_model(), pk_set=new_ids, using=db,
+        )
+
+    def _to_tag_model_instances(self, tags):
+        """
+        Takes an iterable containing either strings, tag objects, or a mixture
+        of both and returns set of tag objects.
+        """
+        db = router.db_for_write(self.through, instance=self.instance)
+
         str_tags = set()
         tag_objs = set()
+
         for t in tags:
             if isinstance(t, self.through.tag_model()):
                 tag_objs.add(t)
             elif isinstance(t, six.string_types):
                 str_tags.add(t)
             else:
-                raise ValueError("Cannot add {0} ({1}). Expected {2} or str.".format(
-                    t, type(t), type(self.through.tag_model())))
+                raise ValueError(
+                    "Cannot add {0} ({1}). Expected {2} or str.".format(
+                        t, type(t), type(self.through.tag_model())))
 
         if getattr(settings, 'TAGGIT_CASE_INSENSITIVE', False):
             # Some databases can do case-insensitive comparison with IN, which
@@ -173,26 +224,30 @@ class _TaggableManager(models.Manager):
 
             for name in str_tags:
                 try:
-                    tag = self.through.tag_model().objects.get(name__iexact=name)
+                    tag = (self.through.tag_model()._default_manager
+                           .using(db)
+                           .get(name__iexact=name))
                     existing.append(tag)
                 except self.through.tag_model().DoesNotExist:
                     tags_to_create.append(name)
         else:
-            # If str_tags has 0 elements Django actually optimizes that to not do a
-            # query.  Malcolm is very smart.
-            existing = self.through.tag_model().objects.filter(
-                name__in=str_tags
-            )
+            # If str_tags has 0 elements Django actually optimizes that to not
+            # do a query.  Malcolm is very smart.
+            existing = (self.through.tag_model()._default_manager
+                        .using(db)
+                        .filter(name__in=str_tags))
 
             tags_to_create = str_tags - set(t.name for t in existing)
 
         tag_objs.update(existing)
 
         for new_tag in tags_to_create:
-            tag_objs.add(self.through.tag_model().objects.create(name=new_tag))
+            tag_objs.add(
+                self.through.tag_model()._default_manager
+                .using(db)
+                .create(name=new_tag))
 
-        for tag in tag_objs:
-            self.through.objects.get_or_create(tag=tag, **self._lookup_kwargs())
+        return tag_objs
 
     @require_instance_manager
     def names(self):
@@ -203,23 +258,91 @@ class _TaggableManager(models.Manager):
         return self.get_queryset().values_list('slug', flat=True)
 
     @require_instance_manager
-    def set(self, *tags):
-        self.clear()
-        self.add(*tags)
+    def set(self, *tags, **kwargs):
+        """
+        Set the object's tags to the given n tags. If the clear kwarg is True
+        then all existing tags are removed (using `.clear()`) and the new tags
+        added. Otherwise, only those tags that are not present in the args are
+        removed and any new tags added.
+        """
+        db = router.db_for_write(self.through, instance=self.instance)
+        clear = kwargs.pop('clear', False)
+
+        if clear:
+            self.clear()
+            self.add(*tags)
+        else:
+            # make sure we're working with a collection of a uniform type
+            objs = self._to_tag_model_instances(tags)
+
+            # get the existing tag strings
+            old_tag_strs = set(self.through._default_manager
+                               .using(db)
+                               .filter(**self._lookup_kwargs())
+                               .values_list('tag__name', flat=True))
+
+            new_objs = []
+            for obj in objs:
+                if obj.name in old_tag_strs:
+                    old_tag_strs.remove(obj.name)
+                else:
+                    new_objs.append(obj)
+
+            self.remove(*old_tag_strs)
+            self.add(*new_objs)
 
     @require_instance_manager
     def remove(self, *tags):
-        self.through.objects.filter(**self._lookup_kwargs()).filter(
-            tag__name__in=tags).delete()
+        if not tags:
+            return
+
+        db = router.db_for_write(self.through, instance=self.instance)
+
+        qs = (self.through._default_manager.using(db)
+              .filter(**self._lookup_kwargs())
+              .filter(tag__name__in=tags))
+
+        old_ids = set(qs.values_list('tag_id', flat=True))
+
+        signals.m2m_changed.send(
+            sender=self.through, action="pre_remove",
+            instance=self.instance, reverse=False,
+            model=self.through.tag_model(), pk_set=old_ids, using=db,
+        )
+        qs.delete()
+        signals.m2m_changed.send(
+            sender=self.through, action="post_remove",
+            instance=self.instance, reverse=False,
+            model=self.through.tag_model(), pk_set=old_ids, using=db,
+        )
 
     @require_instance_manager
     def clear(self):
-        self.through.objects.filter(**self._lookup_kwargs()).delete()
+        db = router.db_for_write(self.through, instance=self.instance)
 
-    def most_common(self):
-        return self.get_queryset().annotate(
+        signals.m2m_changed.send(
+            sender=self.through, action="pre_clear",
+            instance=self.instance, reverse=False,
+            model=self.through.tag_model(), pk_set=None, using=db,
+        )
+
+        self.through._default_manager.using(db).filter(
+            **self._lookup_kwargs()).delete()
+
+        signals.m2m_changed.send(
+            sender=self.through, action="post_clear",
+            instance=self.instance, reverse=False,
+            model=self.through.tag_model(), pk_set=None, using=db,
+        )
+
+    def most_common(self, min_count=None, extra_filters=None):
+        queryset = self.get_queryset(extra_filters).annotate(
             num_times=models.Count(self.through.tag_relname())
         ).order_by('-num_times')
+        if min_count:
+            queryset = queryset.filter(num_times__gte=min_count)
+
+        return queryset
 
     @require_instance_manager
     def similar_objects(self):
@@ -237,12 +360,13 @@ class _TaggableManager(models.Manager):
             # Can we do this without a second query by using a select_related()
             # somehow?
             f = _get_field(self.through, lookup_keys[0])
-            rel_model = f.rel.model if VERSION >= (1, 9) else f.rel.to
+            remote_field = _remote_field(f)
+            rel_model = _related_model(_remote_field(f))
             objs = rel_model._default_manager.filter(**{
-                "%s__in" % f.rel.field_name: [r["content_object"] for r in qs]
+                "%s__in" % remote_field.field_name: [r["content_object"] for r in qs]
             })
             for obj in objs:
-                items[(getattr(obj, f.rel.field_name),)] = obj
+                items[(getattr(obj, remote_field.field_name),)] = obj
         else:
             preload = {}
             for result in qs:
@@ -323,22 +447,17 @@ class TaggableManager(RelatedField, Field):
             del kwargs[kwarg]
         # Add arguments related to relations.
         # Ref: https://github.com/alex/django-taggit/issues/206#issuecomment-37578676
-        if isinstance(self.rel.through, six.string_types):
-            kwargs['through'] = self.rel.through
-        elif not self.rel.through._meta.auto_created:
-            kwargs['through'] = "%s.%s" % (self.rel.through._meta.app_label, self.rel.through._meta.object_name)
+        rel = _remote_field(self)
+        if isinstance(rel.through, six.string_types):
+            kwargs['through'] = rel.through
+        elif not rel.through._meta.auto_created:
+            kwargs['through'] = "%s.%s" % (rel.through._meta.app_label, rel.through._meta.object_name)
 
-        # rel.to renamed to remote_field.model in Django 1.9
-        if VERSION >= (1, 9):
-            if isinstance(self.remote_field.model, six.string_types):
-                kwargs['to'] = self.remote_field.model
-            else:
-                kwargs['to'] = '%s.%s' % (self.remote_field.model._meta.app_label, self.remote_field.model._meta.object_name)
+        related_model = _related_model(rel)
+        if isinstance(related_model, six.string_types):
+            kwargs['to'] = related_model
         else:
-            if isinstance(self.rel.to, six.string_types):
-                kwargs['to'] = self.rel.to
-            else:
-                kwargs['to'] = '%s.%s' % (self.rel.to._meta.app_label, self.rel.to._meta.object_name)
+            kwargs['to'] = '%s.%s' % (related_model._meta.app_label, related_model._meta.object_name)
 
         return name, path, args, kwargs
 
@@ -356,9 +475,11 @@ class TaggableManager(RelatedField, Field):
             # rel.to renamed to remote_field.model in Django 1.9
             if VERSION >= (1, 9):
                 if isinstance(self.remote_field.model, six.string_types):
-                    def resolve_related_class(field, model, cls):
+                    def resolve_related_class(cls, model, field):
                         field.remote_field.model = model
-                    add_lazy_relation(cls, self, self.remote_field.model, resolve_related_class)
+                    lazy_related_operation(
+                        resolve_related_class, cls, self.remote_field.model, field=self
+                    )
             else:
                 if isinstance(self.rel.to, six.string_types):
                     def resolve_related_class(field, model, cls):
@@ -366,13 +487,22 @@ class TaggableManager(RelatedField, Field):
                     add_lazy_relation(cls, self, self.rel.to, resolve_related_class)
 
             if isinstance(self.through, six.string_types):
-                def resolve_related_class(field, model, cls):
-                    self.through = model
-                    self.rel.through = model
-                    self.post_through_setup(cls)
-                add_lazy_relation(
-                    cls, self, self.through, resolve_related_class
-                )
+                if VERSION >= (1, 9):
+                    def resolve_related_class(cls, model, field):
+                        self.through = model
+                        self.remote_field.through = model
+                        self.post_through_setup(cls)
+                    lazy_related_operation(
+                        resolve_related_class, cls, self.through, field=self
+                    )
+                else:
+                    def resolve_related_class(field, model, cls):
+                        self.through = model
+                        _remote_field(self).through = model
+                        self.post_through_setup(cls)
+                    add_lazy_relation(
+                        cls, self, self.through, resolve_related_class
+                    )
             else:
                 self.post_through_setup(cls)
 
@@ -520,7 +650,7 @@ class TaggableManager(RelatedField, Field):
         opts = self.through._meta
         linkfield = _get_field(self.through, self.m2m_reverse_field_name())
         if direct:
-            join1infos = [PathInfo(self.model._meta, opts, [from_field], self.rel, True, False)]
+            join1infos = [PathInfo(self.model._meta, opts, [from_field], _remote_field(self), True, False)]
             join2infos = linkfield.get_path_info()
         else:
             join1infos = linkfield.get_reverse_path_info()
@@ -574,7 +704,7 @@ def _get_subclasses(model):
     for field in all_fields:
         # Django 1.8 +
         if (not RelatedObject and isinstance(field, OneToOneRel) and
-                getattr(field.field.rel, "parent_link", None)):
+                getattr(_remote_field(field.field), "parent_link", None)):
             subclasses.extend(_get_subclasses(field.related_model))
 
         # < Django 1.8
